/************************************************************
**
**       COPYRIGHT (C) 1993 UNIVERSITY OF PITTSBURGH
**       COPYRIGHT (C) 1996 GANNON UNIVERSITY
**                  ALL RIGHTS RESERVED
**
**        This software is distributed on an as-is basis
**        with no warranty implied or intended.  No author
**        or distributor takes responsibility to anyone 
**        regarding its use of or suitability.
**
**        The software may be distributed and modified 
**        freely for academic and other non-commercial
**        use but may NOT be utilized or included in whole
**        or part within any commercial product.
**
**        This copyright notice must remain on all copies 
**        and modified versions of this software.
**
************************************************************/

/* 
 *	file var.c
 * As part of MS Thesis Work, S T Frezza, UPitt Engineering
 * June, 1990
 *
 * This file contains the structure and union manipulation functions that 
 * operate on VAR types and the underlying structures.  These are based on the 
 * following definitions:

 * 	typedef union  var_item_union	varItem;
 * 	typedef struct var_item_struct  varStruct;
 * 	typedef struct var_struct 	var;
 * 	typedef struct var_list_struct  varlist;

 * 	struct var_struct
 * 	{
 *     		int 	action;
 *     		varStruct 	*item1;
 *     		varStruct	*item2;
 * 	};

 * 	struct var_item_struct
 * 	{
 *    		varItem   *this;
 *     		int       type;
 * 	};

 * 	union var_union_type
 * 	{
 *     		int     *i;
 *     		srange  *r;
 *     		var *v;
 * 	};
 
 * 	struct var_struct_list
 * 	{
 *     		var       *this;
 *     		varlist   *next;
 * 	};
 */

/*
 * These structures essentially create an evaluation tree for operations on ranges;  It
 * also implements a delayed evaluation mechanism.  This means 
 * that there is an entire algebra being implemented (at least partially) within the following
 * code.  What is here is not that smart, and only has for operations { +, -, /, * }.

 * For example, the sum of two ranges (caused by the evaluation of a _var_ structure with
 * the operation ADD, and having two _varStruct_s which are both ranges yeilds a single 
 * range, which is the first range having both it's min and max fields augmented by the
 * (integer) result of evaluating the second range.  Note that the second range is altered
 * (reduced) during this operation.
 */

#include "includes\\externs.h"
#include <stdlib.h>


/* ---- Local Variables ---------------------------------------------------------------*/

srange *create_srange(int q1, int q2, int or)
{
    srange *r = (srange *)calloc(1, sizeof(srange));
    r->q1 = MIN(q1, q2);
    r->q2 = MAX(q1, q2);
    r->orient = or;
    return(r);
}


/*-------------------------------------------------------------------------------------*/
int eval_srange(srange *r)
{
    /* This returns an integer equivalent to the center of the given range: */
    int sum;
    if (r->q1 != r->q2)
    {
	sum = r->q1 + r->q2;
	r->q1 = r->q2 = sum/2;
    }
    return(r->q1);
}
	
/*-------------------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------*/
/* end of file "var.c" */
